/* -*- Mode: C++ -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "internal.h"

/* ********************************************************************** */
/* UTF-16 decoder
 * Inspired by above utf-8 decoder, but not optimised for table lookup.
 * This way, the most likely case (namely, a plain UCS-2 char) has the longest
 * path with 3 ifs, which is not optimal.
 */
static INLINE LiqChar32 utf16_decode (LiqChar32 *state, LiqChar32 *codep, LiqChar32 word)
{
    LiqChar32 w10= word >> 10;
    if (w10 == 0x37) { /* low surrogate */
        if ((*state >> 10) == 0) /* no preceding high surrogate? */
            *state= UTF_REJECT;
        else {
            *codep= (*state << 10) + word - 0x35fdc00;
            if ((*codep >> 11) == 0x1b)
                *state= UTF_REJECT; /* illegal: encoded surrogate */
            else
                *state= UTF_ACCEPT;
        }
    }
    else
    if (*state != UTF_ACCEPT)
        *state= UTF_REJECT;
    else
    if (w10 == 0x36) /* high surrogate */
        *state= word;
    else
        *codep= word;
    return *state;
}

static size_t liq_utf16_strnlen (LiqChar16 const *s, size_t n)
{
    LiqChar32 cp;
    LiqChar32 state= UTF_ACCEPT;
    size_t    len= 0;
    while (n--) {
        if (utf16_decode(&state, &cp, *s++) == UTF_ACCEPT)
            len++;
    }
    if (state != UTF_ACCEPT) {
        assert (0 && "bad utf-16 encoding"); /* die in debug mode */
        return 0;                            /* reject string completely in production code */
    }
    return len;
}

void liq_render_set_text_n_utf16 (LiqRender *self, LiqChar16 const *s, size_t n)
{
    liq_render_alloc_info (self, liq_utf16_strnlen (s, n));
    {
        LiqCharInfo *c= self->m_info;
        LiqChar32 cp= 0;
        LiqChar32 state= UTF_ACCEPT;
        while (n--) {
            if (utf16_decode(&state, &cp, *s++) == UTF_ACCEPT)
                liq_set_info (c++, cp);
        }
    }
    liq_render_finalise_text (self);
}

static size_t my_strlen16 (LiqChar16 const *s)
{
    size_t n= 0;
    while (*s++) n++;
    return n;
}

void liq_render_set_text_utf16 (LiqRender *self, LiqChar16 const *text)
{
    liq_render_set_text_n_utf16 (self, text, my_strlen16 (text));
}

